# Создание контейнера

После регистрации в `DIContainerBuilder`, или просто `builder`, библиотека должна проверить, что не нарушены правила зависимостей между типами. После чего она оптимизирует данные для дальнейшего использования, и выдает на выходе контейнер, с которым и будет дальнейшая работа.

Основная цель создания контейнера - отсечь все возможные ошибки как можно раньше. 

## Синтаксис
Чтобы создать контейнер, достаточно у билдера, вызвать метод `build`:
```Swift
let container = try! builder.build()
```

## Процесс создания

Во время создания проверяется, не нарушены ли правила регистрации и нет ли конфликтов в зависимостях. Поэтому метод `build` может кинуть исключение. Для удобства пользования все возникшие ошибки объединяются в одно исключение, содержащее все обнаруженные ошибки.
Настоятельно рекомендую не игнорировать эти исключения, и ронять программу в случае возникновения ошибок, но библиотека не заставляет это делать обязательно, поэтому в случае ошибки приложение может выполнять какие-нибудь специфические действия.

## Проверки и исключения
Все возникшие ошибки объединяются в: `DIError.build(errors: [DIError])`. Далее по тексту будет идти речь об ошибках, которые попадают в список errors.

!! Во время чтения этой главы, будет описан синтаксис, не встречающийся на текущий момент в документации, поэтому рекомендую вернуться еще раз к этой главе позже.

Во время создания контейнера, библиотека проверяет:
* Нет совпадений имен для одного и того же типа.
> Ошибка: **`DIError.intersectionNamesForType(type:, names:, typesInfo:)`**
> Параметры: type - тип, для которого пересекаются имена
>            names - множество имен, которые пересекаются
>            typesInfo - информация о всех регистрациях для этого типа
> Пример:
> ```Swift
> builder.register(type: Cat.self).set(name: "Felix")...
> builder.register(type: Cat.self).set(name: "Felix")...
> ```
> В данном примере, регистрируются две кошки, с одним и тем же именем.
***

* При множественной регистрации указано не более одного default:
> Ошибка: **`DIError.pluralSpecifiedDefaultType(type:, typesInfo:)`**
> Параметры: type - тип, для которого указано несколько регистраций
>            typesInfo - информация о всех регистрациях для этого типа 
> Пример:
> ```Swift
> builder.register(type: Cat.self).set(.default)...
> builder.register(type: Cat.self).set(.default)...
> ```
> В данном примере, регистрируются две кошки, и обе объявлены 'по умолчанию'.
***

* У типа есть хотябы один метод инициализации, если не указано явно обратного.
> Ошибка: **`DIError.notSpecifiedInitializationMethodFor(typeInfo:)`**
> Параметры: typeInfo - информация о регистрации, в которой был упущен метод инициализации
> Пример:
> ```Swift
> builder.register(type: Cat.self)
> try! builder.build()
> ```
> В данном примере, регистрируется кошка, но не указывается, как её создавать. Такая регистрация не имеет смысла, так как в дальнейшем не получиться создать объявленный тип.
***


#### Примечания
##### 
typeInfo - это специальный тип DITypeInfo, который содержит более точную информацию, нежели тип. 
В typeInfo содержится:
* type - тип, для которого проходила регистрация
* file - файл, в котором проходила регистрация
* line - строчка кода, в которой проходила регистрация
Таким образом, можно легко найти в каком месте была произведена ошибка.

#####
```Swift
builder.register(type: Cat.self).as(Birds.self).unsafe()...
``` 
Данный пример содержит ошибку: кот не является птицей, но, не смотря на это, библиотека пропустит такой код на стадии компиляции и на стадии сборки контейнера, так как явно указано - не проверять тип.
Ошибка обнаружиться только в тот момент, когда программа запросит у библиотеки тип `Birds` - будет произведена проверка, что кошка не является птицей и кинется исключение: 

> Ошибка: **`DIError.typeIsIncorrect(requestedType:, realType:, typeInfo:)`**  
> Параметры:  
>            requestedType - запрашиваемый тип  
>            realType - тип объекта, который был создан  
>            typeInfo - информация о регистрации
***

#### [Главная](main.md)
#### [Предыдущая глава "Регистрация"](registration.md)
#### [Следующая глава "Разрешение зависимостей"](resolve.md)
